/**
 * SmartCGMS - continuous glucose monitoring and controlling framework
 * https://diabetes.zcu.cz/
 *
 * Copyright (c) since 2018 University of West Bohemia.
 *
 * Contact:
 * diabetes@mail.kiv.zcu.cz
 * Medical Informatics, Department of Computer Science and Engineering
 * Faculty of Applied Sciences, University of West Bohemia
 * Univerzitni 8, 301 00 Pilsen
 * Czech Republic
 * 
 * 
 * Purpose of this software:
 * This software is intended to demonstrate work of the diabetes.zcu.cz research
 * group to other scientists, to complement our published papers. It is strictly
 * prohibited to use this software for diagnosis or treatment of any medical condition,
 * without obtaining all required approvals from respective regulatory bodies.
 *
 * Especially, a diabetic patient is warned that unauthorized use of this software
 * may result into severe injure, including death.
 *
 *
 * Licensing terms:
 * Unless required by applicable law or agreed to in writing, software
 * distributed under these license terms is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * a) For non-profit, academic research, this software is available under the
 *      GPLv3 license.
 * b) For any other use, especially commercial use, you must contact us and
 *       obtain specific terms and conditions for the use of the software.
 * c) When publishing work with results obtained using this software, you agree to cite the following paper:
 *       Tomas Koutny and Martin Ubl, "Parallel software architecture for the next generation of glucose
 *       monitoring", Procedia Computer Science, Volume 141C, pp. 279-286, 2018
 */

//#define DRUN_AS_DEBUG

#include "execute_model.h"

#include <rtl/referencedImpl.h>
#include <rtl/FilterLib.h>
#include <rtl/scgmsLib.h>

#include <array>
#include <numeric>
#include <cmath>

#ifdef DRUN_AS_DEBUG
	#include <cstdlib>
	#include <iostream>
#endif

struct TConfiguration {
	const char* config;
	const GUID signal_id;		//signal to capture
};

const char* Icarus_V1_config = R"CLC(
; Signal generator
[Filter_001_{9EEB3451-2A9D-49C1-BA37-2EC0B00E5E6D}]
Model = {8114B2A6-B4B2-4C8D-A029-625CBDB682EF}
Feedback_Name = 
Synchronize_To_Signal = true
Synchronization_Signal = {FFFFFFFF-FFFF-FFFF-FFFF-FFFFFFFFFFFF} 
Stepping = 0.00347222222222222222222222222222
Maximum_Time = 0
Shutdown_After_Last = false
Time_Segment_Id = 1
Parameters = 0.005 0.005 0 0.1 0.001 0.001 8 10 0.05 0.01 0.01 0.01 0 0 0 0 0 0 0 0 0 0 0 0 -0.5 -10 0 -1 0 0.05 0.028735 0.028344 5e-05 0.3 0 0 12 70 0.22 0.05 0.05 0.04 209 9.2 220 0 0 0 0 0 0 95 0 0.929 -0.037 1.308 0.003472 0 0 0.35 0.1 0.1 0.05 1 0.3 0.3 18 100 1 1 1 1 200 20 300 300 100 200 150 150 50 300 5 2 0 5 0.041667 0 0.041667 1.5
)CLC";

constexpr GUID signal_Bergman_BG = { 0xe1cd07ef, 0xb079, 0x4911, { 0xb7, 0x9b, 0xd2, 0x3, 0x48, 0x61, 0x1, 0xc8 } };

const std::array<TConfiguration, 1> available_configurations = { {Icarus_V1_config, signal_Bergman_BG} };


#pragma warning( push )
#pragma warning( disable : 4250 ) // C4250 - 'class1' : inherits 'class2::member' via dominance
 
class CSignal_Terminal : public virtual scgms::IFilter, public virtual refcnt::CNotReferenced {
protected:
	const GUID mSignal_Id = Invalid_GUID;
	double *mData_Ptr = nullptr;
	const double *mData_End = nullptr;
public:	
	CSignal_Terminal(const GUID signal_id, const double *data, const size_t count) :
		mSignal_Id(signal_id), mData_Ptr(const_cast<double*>(data)), mData_End(data + count) {};

	virtual ~CSignal_Terminal() {
		//sanitize the end of data 
		while (mData_Ptr < mData_End) {
			*mData_Ptr = std::numeric_limits<double>::quiet_NaN();
			mData_Ptr++;
		}
	}
public:
	virtual HRESULT Execute(scgms::IDevice_Event *raw_event) override final {
		scgms::UDevice_Event event{ raw_event };

		if ((event.event_code() == scgms::NDevice_Event_Code::Level) &&
			(event.signal_id() == mSignal_Id) &&
			(mData_Ptr < mData_End)) {

			*mData_Ptr = event.level();
			mData_Ptr++;
		}

		//event is relased as it goes out of scope
		return S_OK;
	};
	virtual HRESULT Configure(scgms::IFilter_Configuration* configuration, refcnt::wstr_list *error_description) override final {
		return S_OK;	//nothing to configure
	}

};


#pragma warning( pop )


int execute_model(const uint8_t configuration_id, const double *cho, const double *insulin_bolus, const double *insulin_basal_rate, double const *bg, const size_t count) {
	if (!scgms::is_scgms_loaded()) return static_cast<int>(CO_E_ERRORINDLL);
	

	const uint8_t effective_configuration_id = configuration_id - 1;
	if (effective_configuration_id >= static_cast<decltype(effective_configuration_id)>(NExecute_Model_Id::count)) return static_cast<int>(E_INVALIDARG);

	const TConfiguration& config = available_configurations[effective_configuration_id];

	//1. setup the configuration
#ifdef  DRUN_AS_DEBUG
	printf("Selected configuration:\n %s", config.config);
#endif

	refcnt::Swstr_list errors = refcnt::Swstr_list{};
	scgms::SPersistent_Filter_Chain_Configuration configuration{};
	HRESULT rc = configuration->Load_From_Memory(config.config, strlen(config.config), errors.get());
	if (rc == S_OK) {
		//2. create filter chain executor
		CSignal_Terminal signal_terminal{ config.signal_id, bg, count };		
		{	//3. create a sub-block to make the signal_terminal outlive the executor
			scgms::SFilter_Executor executor(configuration.get(), nullptr, nullptr, errors, &signal_terminal);

			double simulation_time = 0.0;

			if (executor.operator bool()) {
				//3. feed the executor with the events
				//this will fill the output array via signal_terminal
				auto send_event = [&executor, &rc, &simulation_time](const GUID &signal_id, const double level, 
					const scgms::NDevice_Event_Code code = scgms::NDevice_Event_Code::Level, const uint64_t segment_id = 1)->HRESULT {
					constexpr GUID device_id = { 0x49258a22, 0xcf53, 0x4bc1, { 0x92, 0xfb, 0x9a, 0x2a, 0x8f, 0xad, 0x52, 0xbf } }; // {49258A22-CF53-4BC1-92FB-9A2A8FAD52BF}

					scgms::UDevice_Event event{ code};
					event.level() = level;
					event.signal_id() = signal_id;
					event.device_id() = device_id;
					event.device_time() = simulation_time;
					event.segment_id() = segment_id;

					return executor.Execute(std::move(event));
				};

				rc = send_event(scgms::signal_All, 0.0, scgms::NDevice_Event_Code::Time_Segment_Start);
				double last_basal_rate = -1.0;	//physically impossible value
				for (size_t i = 0; (i < count) && (SUCCEEDED(rc)); i++) {
					auto send_level = [&send_event, i, &rc](const GUID &signal_id, const double level) {
						if (SUCCEEDED(rc)) {
							if (std::isnormal(level) && (level > 0.0)) rc = send_event(signal_id, level);
						}
					};

					send_level(scgms::signal_Carb_Intake, cho[i]);
					send_level(scgms::signal_Requested_Insulin_Bolus, insulin_bolus[i]);
					if (insulin_basal_rate[i] != last_basal_rate) {
						send_level(scgms::signal_Requested_Insulin_Basal_Rate, insulin_basal_rate[i]);
						last_basal_rate = insulin_basal_rate[i];
					}
					
					simulation_time += 5.0*scgms::One_Minute;
				}	
				auto unused1 = send_event(scgms::signal_Null, 0.0);	//advance the model to process the recently sent levels
				auto unused2 = send_event(scgms::signal_All, 0.0, scgms::NDevice_Event_Code::Time_Segment_Stop);

				//4. gracefully terminate the entire chain by processing the shutdown event as well
				rc = send_event(scgms::signal_All, 0.0, scgms::NDevice_Event_Code::Shut_Down, scgms::All_Segments_Id);
				executor->Terminate(SUCCEEDED(rc) ? TRUE : FALSE);
			}
			else {
#ifdef  DRUN_AS_DEBUG
				printf("Failed to execute the configuration!\n");
#endif
				rc = E_FAIL;
			}
		}
	}
	else {
#ifdef  DRUN_AS_DEBUG
		printf("Configuration failed!\n");
#endif
	}

#ifdef  DRUN_AS_DEBUG
	refcnt::wstr_container *wstr;
	if (errors->empty() != S_OK) {
        	std::wcerr << "Error description: " << std::endl;
	        while (errors->pop(&wstr) == S_OK) {
        	    std::wcerr << refcnt::WChar_Container_To_WString(wstr) << std::endl;
	            wstr->Release();
	        }
	    }
#endif

	return static_cast<int>(rc);
}


#ifdef _WIN32

	void __cdecl test_execute_model(HWND hwndStub, HINSTANCE  hAppInstance, LPCTSTR lpCmdLine, int nCmdShow) {
		//Actually, this is a test function
		constexpr size_t count = 602;
		std::vector<double> input(count), output(count);

		std::iota(input.begin(), input.end(), 0.0);

		execute_model(1, input.data(), input.data(), input.data(), output.data(), input.size());
	}

#else 
	#ifdef DRUN_AS_DEBUG
	int main(int argc, char** argv) {
		constexpr size_t count = 602;
		std::vector<double> input(count), output(count);

		std::iota(input.begin(), input.end(), 0.0);

		int rc = execute_model(1, input.data(), input.data(), input.data(), output.data(), input.size());
		printf("Return code: 0x%x\nFirst output value: %f\n", rc, output[0]);

		return rc;
	}
	#endif
#endif
